#!/usr/bin/env python
# coding: utf-8
# Copyright © 2015 Wieland Hoffmann
# License: MIT, see LICENSE for details
import enum


class State(enum.Enum):
    active = 1
    reloading = 2
    inactive = 3
    failed = 4
    activating = 5
    deactivating = 6
    unknown = 7



#!/usr/bin/env python
# coding: utf-8
# Copyright © 2015 Wieland Hoffmann
# License: MIT, see LICENSE for details
from argparse import Action


class TestAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if not namespace.unit:
            # Not only does this mean we error out on missing --unit arguments,
            # but also on --all-units because --unit conflicts with it
            parser.error("--test needs at least one --unit")
        else:
            setattr(namespace, self.dest, True)



#!/usr/bin/env python
# coding: utf-8
# Copyright © 2015 Wieland Hoffmann
# License: MIT, see LICENSE for details
from ..notifier import INotifier
from .. import state_helpers
from email.mime.text import MIMEText
from StringIO import StringIO
from twisted.mail.smtp import (messageid, sendmail)
from twisted.plugin import IPlugin
from zope.interface.declarations import implementer


@implementer(IPlugin, INotifier)
class SMTPNotifier(object):
    name = "smtp"
    description = "Send notification emails via twisted.mail.smtp.sendmail"

    def add_arguments(self, group):
        group.add_argument("--smtp-from", action="store",
                           help="The address to use in the From: header")
        group.add_argument("--smtp-to", action="store",
                           help="The address to use in the To: header")
        group.add_argument("--smtp-user", action="store", default=None,
                           help="The SMTP username")
        group.add_argument("--smtp-password", action="store", default=None,
                           help="The SMTP password")
        group.add_argument("--smtp-host", action="store",
                           help="The SMTP host")
        group.add_argument("--smtp-port", action="store", type=int,
                           help="The SMTP port")
        group.add_argument("--smtp-require-authentication", action="store_true",
                           default=False, help="Require authentication")
        group.add_argument("--smtp-require-transport-security",
                           action="store_true", default=False,
                           help="Require STARTTLS")

    def handle_arguments(self, args):
        self.from_ = args.smtp_from
        self.to = args.smtp_to
        self.user = args.smtp_user
        self.password = args.smtp_password
        self.host = args.smtp_host
        self.port = args.smtp_port
        self.auth = args.smtp_require_authentication
        self.transport_sec = args.smtp_require_transport_security

    def _build_message_file(self, msg):
        """
        :type msg: str
        """
        msg = msg
        message = MIMEText(msg, _subtype='plain', _charset='utf-8')
        message['Subject'] = "sagbescheid service notification"
        message['From'] = self.from_
        message['To'] = self.to
        message['Message-ID'] = messageid()
        return StringIO(message.as_string())

    def _send_mail(self, msg):
        """
        :type msg: str
        """
        message = self._build_message_file(msg)
        sendmail(self.host, self.from_, [self.to], message, port=self.port,
                 username=self.user, password=self.password,
                 requireAuthentication=self.auth,
                 requireTransportSecurity=self.transport_sec)

    def state_changed(self, unit, old_state, new_state):
        """
        :type unit: str
        :type old_state: :class:`sagbescheid.state.State`
        :type new_state: :class:`sagbescheid.state.State`
        """
        if state_helpers.is_ongoing_failure(old_state, new_state):
            self._send_mail("%s is still failing." % unit)
        elif state_helpers.is_failure(new_state):
            self._send_mail("%s entered failed state." % unit)
        elif state_helpers.is_recovery(old_state, new_state):
            self._send_mail("%s recovered." % unit)

sendmailnotifier = SMTPNotifier()



#!/usr/bin/env python
# coding: utf-8
# Copyright © 2015 Wieland Hoffmann
# License: MIT, see LICENSE for details
from ..notifier import INotifier
from ..version import __version__
from .. import state_helpers
from twisted.internet import protocol, reactor
from twisted.plugin import IPlugin
from twisted.words.protocols.irc import IRCClient
from zope.interface.declarations import implementer


class IRCNotifierBot(IRCClient):
    versionName = "sagbescheid"
    versionNum = __version__
    lineRate = 1

    @property
    def nickname(self):
        return self.factory.nick

    def signedOn(self):
        self.join(self.factory.channel)

    def _msg_channel(self, msg):
        """Send ``msg`` to the configured channel.

        :type msg: str
        """
        self.msg(self.factory.channel, msg)

    def state_changed(self, unit, old_state, new_state):
        """
        :type unit: str
        :type old_state: :class:`sagbescheid.state.State`
        :type new_state: :class:`sagbescheid.state.State`
        """
        unit = unit.encode("utf-8")
        if state_helpers.is_ongoing_failure(old_state, new_state):
            self._msg_channel("%s is still failing." % unit)
        elif state_helpers.is_failure(new_state):
            self._msg_channel("%s entered failed state." % unit)
        elif state_helpers.is_recovery(old_state, new_state):
            self._msg_channel("%s recovered." % unit)
        elif state_helpers.is_normal_stop(old_state, new_state):
            self._msg_channel("%s stopped normally" % unit)
        elif state_helpers.is_normal_start(old_state, new_state):
            self._msg_channel("%s started normally" % unit)


@implementer(IPlugin, INotifier)
class IRCNotifierFactory(protocol.ReconnectingClientFactory):
    name = "irc"
    description = "Log events to an IRC channel"

    protocol = IRCNotifierBot

    def add_arguments(self, group):
        group.add_argument("--irc-nick", action="store",
                           help="Nick for the bot")
        group.add_argument("--irc-channel", action="store",
                           help="Channel for the bot to join")
        group.add_argument("--irc-server", action="store",
                           help="IRC server address")
        group.add_argument("--irc-port", action="store", type=int, default=6667,
                           help="IRC server port")

    def handle_arguments(self, args):
        self.channel = args.irc_channel
        self.nick = args.irc_nick
        self.port = args.irc_port
        self.server = args.irc_server
        reactor.connectTCP(self.server, self.port, self)

    def state_changed(self, *args):
        """Pass state change events through to the client.
        """
        self.prot.state_changed(*args)

    def buildProtocol(self, addr):
        self.prot = protocol.ReconnectingClientFactory.buildProtocol(self, addr)
        return self.prot

obj = IRCNotifierFactory()



#!/usr/bin/env python
# coding: utf-8
# Copyright © 2015 Wieland Hoffmann
# License: MIT, see LICENSE for details
import logging


from .state import State
from . import state_helpers
from io import BytesIO
from xml.etree import ElementTree
from twisted.internet import defer


SERVICE_IFACE = "org.freedesktop.systemd1.Service"
UNIT_IFACE = "org.freedesktop.systemd1.Unit"


PATH_REPLACEMENTS = {
    ".": "_2e",
    "-": "_2d",
    "/": "_2f"
}

SYSTEMD_BUS_NAME = "org.freedesktop.systemd1"
UNIT_PATH_PREFIX = "/org/freedesktop/systemd1/unit/"


def make_path(unit):
    for from_, to in PATH_REPLACEMENTS.iteritems():
        unit = unit.replace(from_, to)
    return unit


class Unit(object):
    def __init__(self, name, notifier_registry):
        """
        :type name: str
        :type notifier_registry: :class:`sagbescheid.notifier.NotifierRegistry`
        """
        self.object_path = name
        self.notifier_registry = notifier_registry
        self.state = State.unknown

    @classmethod
    def from_unit_filename(cls, name, notifier_registry):
        """Return an object of class ``cls`` (in the basic case a
        :class:`sagbescheid.unit.Unit` object). The object path will be built
        from the unit filename ``name``.

        :type name: str
        :type notifier_registry: :class:`sagbescheid.notifier.NotifierRegistry`
        """
        name = UNIT_PATH_PREFIX + make_path(name)
        return cls(name, notifier_registry)

    @classmethod
    def from_child_object_path(cls, name, notifier_registry):
        """Return an object of class ``cls`` (in the basic case a
        :class:`sagbescheid.unit.Unit` object). The object path will be built
        from the node name returned by introspecting systemd DBus bus.

        :type name: str
        :type notifier_registry: :class:`sagbescheid.notifier.NotifierRegistry`
        """
        return cls(name, notifier_registry)

    @defer.inlineCallbacks
    def connect(self, con):
        """Connect to the units ``PropertiesChanged`` signal on ``con``.

        :type con: :class:`txdbus.client.DBusClientConnection`
        """
        logging.debug("Connecting %s", self.object_path)
        robj = yield con.getRemoteObject(SYSTEMD_BUS_NAME, self.object_path)
        state = yield robj.callRemote("Get", UNIT_IFACE, "ActiveState")
        logging.debug("The state of %s is %s", self.object_path, state)
        self.state = State[state]
        robj.notifyOnSignal("PropertiesChanged", self.onSignal)

    def onSignal(self, iface, changed, invalidated):
        if iface == UNIT_IFACE:
            new_raw_state = changed.get("ActiveState", None)

            if new_raw_state is None:
                return

            logging.debug("%s: state change to %s, old was %s",
                          self.object_path, new_raw_state, self.state)
            new_state = State[new_raw_state]
            self.notifier_registry.state_changed(self.object_path, self.state,
                                                 new_state)
            # For now, only toggle the state between active, inactive and failed
            if (state_helpers.is_failure(new_state) or
                state_helpers.is_recovery(self.state, new_state) or
                state_helpers.is_normal_start(self.state, new_state) or
                state_helpers.is_normal_stop(self.state, new_state) or
                state_helpers.is_change_from_unknown(self.state, new_state)):
                self.state = new_state


@defer.inlineCallbacks
def get_all_unit_paths(con):
    robj = yield con.getRemoteObject(SYSTEMD_BUS_NAME, "/org/freedesktop/systemd1")
    dbus_units = yield robj.callRemote("ListUnits",
                                       interface="org.freedesktop.systemd1.Manager")
    units = []
    for elem in dbus_units:
        unit_name = elem[6]
        logging.info("Discovered a new unit at %s", unit_name)
        units.append(unit_name)

    defer.returnValue(units)



#!/usr/bin/env python2
# coding: utf-8
# Copyright © 2015, Wieland Hoffmann
# License: MIT, see LICENSE for details
import argparse
import logging

from .argparse_ext import TestAction
from .notifier import get_all_notifiers, get_enabled_notifiers, NotifierRegistry
from .state import State
from .unit import get_all_unit_paths, Unit, UNIT_IFACE
from functools import partial
from operator import attrgetter
from twisted.internet import defer, reactor
from twisted.python import log
from txdbus import client, error


@defer.inlineCallbacks
def setup(args):
    """
    :type args: :class:`argparse.Namespace`
    """
    con = yield client.connect(reactor, "system")
    registry = NotifierRegistry(get_enabled_notifiers(args.notifier))
    try:
        if args.all_units:
            # Get the names of all units
            units = yield get_all_unit_paths(con)
            for unit in units:
                yield Unit.from_child_object_path(unit, registry).connect(con)
        else:
            for unit in args.unit:
                Unit.from_unit_filename(unit, registry).connect(con)
    except error.DBusException:
        logging.exception(
            "The following exception occured during the initial setup:")
        reactor.stop()


def test(args):
    registry = NotifierRegistry(get_enabled_notifiers(args.notifier))
    units = []
    for unit in args.unit:
        units.append(Unit.from_unit_filename(unit, registry))

    def emit_signals():
        for unit in units:
            unit.state = State[args.test_state_from]
            change = {"ActiveState": args.test_state_to}
            unit.onSignal(UNIT_IFACE, change, None)
            reactor.callLater(2, reactor.stop)

    reactor.callLater(2, emit_signals)


def build_arg_parser():
    parser = argparse.ArgumentParser(prog='sagbescheid',
                                     fromfile_prefix_chars='@',
                                     description='Monitor systemd unit states',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)  # noqa
    available_notifiers = list(get_all_notifiers())
    available_notifier_names = map(attrgetter("name"),
                                   available_notifiers)
    parser.add_argument("--notifier", action="append", default=[],
                        choices=available_notifier_names,
                        help="A notifier to enable.")
    parser.add_argument("-v", "--verbose", action="store_true", default=False,
                        help="Be more verbose.")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--unit", action="append",
                       help="A unit to monitor.")
    group.add_argument("--all-units", action="store_true", default=False,
                       help="Monitor all units.")

    for notifier in available_notifiers:
        arg_group = parser.add_argument_group(notifier.name,
                                              "Arguments for the %s notifier" %
                                              notifier.name)
        notifier.add_arguments(arg_group)
    test_group = parser.add_argument_group("Test notifications", """Send a test
    notification for enabled units""")
    test_group.add_argument("--test",
                            action=TestAction,
                            default=False,
                            nargs=0,
                            help="Enable testing mode")
    test_group.add_argument("--test-state-from", action="store",
                            choices=State.__members__.keys(),
                            default="failed",
                            help="The start state for all units")
    test_group.add_argument("--test-state-to", action="store",
                            choices=State.__members__.keys(),
                            default="active",
                            help="The state units will transition to")
    return parser


def main():
    observer = log.PythonLoggingObserver(loggerName="")
    observer.start()

    parser = build_arg_parser()
    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
    for notifier in get_enabled_notifiers(args.notifier):
        notifier.handle_arguments(args)

    if not args.test:
        reactor.callWhenRunning(partial(setup, args))
    else:
        test(args)
    reactor.run()

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding: utf-8
# Copyright © 2015 Wieland Hoffmann
# License: MIT, see LICENSE for details
import logging


from ..notifier import INotifier
from .. import state_helpers
from twisted.plugin import IPlugin
from zope.interface.declarations import implementer


@implementer(IPlugin, INotifier)
class LoggingNotifier(object):
    name = "logging"
    description = "Log events through Pythons logging module"

    def add_arguments(self, group):
        group.add_argument("--logging-level",
                           action="store",
                           default="info",
                           help="The log level to use",
                           choices=["CRITICAL", "ERROR", "WARNING", "INFO",
                                    "DEBUG"])

    def handle_arguments(self, args):
        level = getattr(logging, args.logging_level.upper())
        logging.basicConfig(level=level)

    def state_changed(self, unit, old_state, new_state):
        """
        :type unit: str
        :type old_state: :class:`sagbescheid.state.State`
        :type new_state: :class:`sagbescheid.state.State`
        """
        if state_helpers.is_ongoing_failure(old_state, new_state):
            logging.info("%s is still failing.", unit)
        elif state_helpers.is_failure(new_state):
            logging.info("%s entered failed state.", unit)
        elif state_helpers.is_recovery(old_state, new_state):
            logging.info("%s recovered.", unit)
        else:
            logging.info("%s changed from %s to %s", unit, old_state,
                         new_state)

obj = LoggingNotifier()



#!/usr/bin/env python
# coding: utf-8
# Copyright © 2015 Wieland Hoffmann
# License: MIT, see LICENSE for details
from . import notifiers
from twisted.plugin import getPlugins
from zope.interface import Attribute, Interface


class INotifier(Interface):
    name = Attribute("The name of the notifier")
    description = Attribute("A description of the notifier")

    def add_arguments(parser):
        """Allows notifiers to add additional command line arguments.

        :type parser: :class:`argparse.ArgumentParser`
        """

    def handle_arguments(args):
        """Called when command line arguments have been parsed.

        :type: :class:`argparse.Namespace`
        """

    def state_changed(unit, old_state, new_state):
        """
        :type unit: str
        :type old_state: :class:`sagbescheid.state.State`
        :type new_state: :class:`sagbescheid.state.State`
        """


class NotifierRegistry(object):
    def __init__(self, notifiers):
        """
        :type notifiers: [:class:`sagbescheid.notifier.INotifier`]
        """
        self.notifiers = {}
        for notifier in notifiers:
            self.notifiers[notifier.name] = notifier

    def state_changed(self, unit, old_state, new_state):
        for notifier in self.notifiers.values():
            notifier.state_changed(unit, old_state, new_state)


def get_all_notifiers():
    return getPlugins(INotifier, notifiers)


def get_enabled_notifiers(enabled_notifier_names):
    """
    :type enabled_notifier_names: :class:`[str]`
    :rtype: :class:`sagbescheid.notifier.INotifier`
    """
    available_notifiers = get_all_notifiers()
    available_notifier_name_map = {n.name: n for n in available_notifiers}
    notifiers_to_activate = []

    for notifier in enabled_notifier_names:
        notifier_plugin = available_notifier_name_map[notifier]
        notifiers_to_activate.append(notifier_plugin)

    return notifiers_to_activate



#!/usr/bin/env python2
from setuptools import setup


setup(name="sagbescheid",
      author="Wieland Hoffmann",
      author_email="themineo@gmail.com",
      packages=["sagbescheid", "sagbescheid.notifiers"],
      package_dir={"sagbescheid": "sagbescheid"},
      download_url=["https://github.com/mineo/sagbescheid/tarball/master"],
      url=["http://github.com/mineo/sagbescheid"],
      license="MIT",
      classifiers=["Development Status :: 4 - Beta",
                   "License :: OSI Approved :: MIT License",
                   "Natural Language :: English",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python :: 2.7"],
      description="",
      long_description=open("README.rst").read(),
      install_requires=["Twisted[tls]>=15.2.0",
                        "txdbus==1.0.11",
                        "zope.interface==4.1.2",
                        "enum34==1.0.4"],
      setup_requires=["setuptools_scm"],
      use_scm_version={"write_to": "sagbescheid/version.py"},
      extras_require={
          'docs': ['sphinx', 'sphinxcontrib-autoprogram']
          }
      )



#!/usr/bin/env python
# coding: utf-8
# Copyright © 2015 Wieland Hoffmann
# License: MIT, see LICENSE for details
from .state import State


def is_failure(state):
    return state == State.failed


def is_normal_start(old, new):
    return old in (State.activating, State.inactive) and new == State.active


def is_normal_stop(old, new):
    return old != State.failed and new == State.inactive


def is_ongoing_failure(old, new):
    return old == State.failed and new == State.failed


def is_recovery(old, new):
    return old == State.failed and new == State.active


def is_change_from_unknown(old, new):
    return old == State.unknown and (new == State.inactive or
                                     new == State.active or
                                     new == State.failed)
